/* Mednafen Network Play Server
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <netdb.h>
#include <errno.h>  
#include <fcntl.h>
#include <stdarg.h>

#include <assert.h>

#include <exception>

#include "types.h"
#include "md5.h"
#include "time64.h"

#ifndef SOL_TCP
#define 	SOL_TCP   IPPROTO_TCP
#endif

static const int MaxNickLen = 150; 		// In bytes, not glyphs(worst-case max glyphs using UTF8 is probably around MaxNickLen / 5).
static const int MaxClientsPerGame = 32;	// Maximum number of clients per game.
static const int MaxControllersPerGame = 8;	// Maximum number of controllers per game.  Due to current protocol limitations, setting this higher than
						// 8 is invalid.

static const int MaxTotalControllersDataSize = 	512; // Maximum data size for all the controller data per game.
						// Assume usable-worst-case 12 bytes per controller(in case of mouse, x delta, y delta, buttons), a maximum of 8
						// controllers, would give us 96 bytes.
						// We'll be extra-generous and set it to 512 bytes(in case anyone is goofy enough to try to use
						// the Family Keyboard over netplay or something).

static const uint32 MaxCommandPayload = 4 * 1024 * 1024;	// In bytes.  The maximum data in the payload of 
							// a command to be received by the server(including save state transfers).  Per client limit.

static const uint32 MinSendQSize = 262144;
static const uint32 MaxSendQSize = 6 * 1024 * 1024;	// Maximum size the internal soft send queue is allowed to grow to.  If it grows
							// larger than this, the connection will be dropped.  Per client limit.

// TODO: Ignoring incoming save state receives for a game if a previous save state transfer is in progress.  Or perhaps, if any sendq is >= MaxCommandPayload
// for the game, ignore all large incoming data transfers.

// TODO: Save state transfer throttling.

// TODO: Text throttling.

// TODO: Connection throttling.

// TODO: Nick change throttling.

#define MDFNNPCMD_SETFPS        0x40 /* Sent from client to server ONLY(it should be ignored server-side if it's not from the first
                                        active player for the game). */
#define MDFNNPCMD_TEXT          0x90
#define MDFNNPCMD_SERVERTEXT    0x93 // Server text message(informational), server->client
#define MDFNNPCMD_ECHO          0x94 // Echos the string(no larger than 256 bytes) back to the client(used for pinging).
#define MDFNNPCMD_INTEGRITY     0x95 // Send from a client to a server, then from the server to all clients.
#define MDFNNPCMD_INTEGRITY_RES 0x96 // Integrity result, sent from the clients to the server.  The result should be no larger
                                     // than 256 bytes.
#define MDFNNPCMD_SETNICK       0x98 /* Sent from client to server only. */
#define MDFNNPCMD_PLAYERJOINED  0xA0    // Data:  <byte: bitmask, which inputs this player controls>
                                        //        <bytestream: nickname>
#define MDFNNPCMD_PLAYERLEFT    0xA1    // Data: (see above)
#define MDFNNPCMD_YOUJOINED     0xB0
#define MDFNNPCMD_YOULEFT       0xB1

#define MDFNNPCMD_NICKCHANGED   0xB8    // Sent from server to client

typedef struct {
	uint32 id;	/* mainly for faster referencing when pointed to from the Games
		           entries.
			 */
	char *nickname;	//[MaxNickLen + 1];

	int TCPSocket;
	void *game;		/* Pointer to the game this player is in. */
	int localplayers;	/* The number of local players, 1 through total_controllers */

	time_t timeconnect;	/* Time the client made the connection. */


	uint8 pending_command;	/* Only for protocol >= 1 */
	int total_controllers;  /* Total controllers for the emulated system. */
	int *controller_data_size; /* Controller data size. */
	int total_controllers_data_size; /* Data size of all controllers */
	int local_controllers_data_size; /* Data size of local controllers */
	int *controller_data_offset;
	int merge_target;	/* 0 if not active, 1 = player 1, 2 = player 2, etc. */

	int protocol_version;	/* 0 = FCE Ultra + Mednafen < 0.5.0 - NOT SUPPORTED*/
				/* 1 = Mednafen >= 0.5.0 
				    Changes:  
					1 extra byte in incoming data packet to denote command, instead
					of using 0xFF in the first joystick byte as an escape.
				   2 = Mednafen >= 0.7.0
				    Changes:
					Support for a wider variety of input devices.
				*/

	/* Variables to handle non-blocking TCP reads. */
	uint8 *nbtcp;
	uint32 nbtcphas, nbtcplen;
	uint32 nbtcptype;

	uint8 *sendq;
	uint32 sendqlen;
	uint32 sendqalloced;
} ClientEntry;

typedef struct
{
	/* Unique 128-bit identifier for this game session. */
        uint8 id[16];

	/* The version of the protocol the clients in this game are communicating with. */
	int ProtocolVersion;

	/* Total controllers for this game. */
	int TotalControllers;

	/* Data size for each controller. */
	int ControllerDataSize[MaxControllersPerGame];

	/* Total data size for all controllers. */
        int TotalControllersDataSize;

	/* Offset into joybuf for each controller. */
	int ControllerDataOffset[MaxControllersPerGame];

        ClientEntry **Clients;  /* Pointers to connected clients. */

	int IsUnique[MaxClientsPerGame];
		 		/* Set to 1 if player is unique client, 0
				   if it's a duplicated entry to handle multiple players
				   per client.
				*/
	int64 last_time;
	uint32 fps;

        uint8 joybuf[MaxTotalControllersDataSize + 1]; /* X player data + 1 command byte8 */
} GameEntry;

typedef struct
{
	int MaxClients;		/* The maximum number of clients to allow. */
				/* You should expect each client to use
				   65-70Kbps(not including save state loads) while
				   connected(and logged in).
				*/

	int ConnectTimeout;	/* How long to wait(in seconds) for the client to provide
				   login data before disconnecting the client.
				*/

	int Port;		/* The port to listen on. */
	uint8 *Password;	/* The server password. */
} CONFIG;

CONFIG ServerConfig;

int LoadConfig(char *fn)
{
 FILE *fp;

 ServerConfig.Port = -1;
 ServerConfig.MaxClients = -1;
 ServerConfig.ConnectTimeout = -1;

 if((fp=fopen(fn,"rb")))
 {
  char buf[256];
  while(fgets(buf, 256, fp) > 0)
  {
   if(!strncasecmp(buf,"maxclients",strlen("maxclients")))
    sscanf(buf,"%*s %d",&ServerConfig.MaxClients);
   else if(!strncasecmp(buf,"connecttimeout",strlen("connecttimeout")))
    sscanf(buf,"%*s %d",&ServerConfig.ConnectTimeout);
   else if(!strncasecmp(buf,"port",strlen("port")))
    sscanf(buf,"%*s %d",&ServerConfig.Port);
   else if(!strncasecmp(buf,"password",strlen("password")))
   {
    char *pass = 0;
    sscanf(buf,"%*s %as",&pass);
    if(pass)
    {
     struct md5_context md5;
     ServerConfig.Password = (uint8 *)malloc(16);
     md5_starts(&md5);
     md5_update(&md5,(uint8*)pass,strlen(pass));
     md5_finish(&md5,ServerConfig.Password);
     puts("Password required to log in.");
    }
   }
  }
 }
 else return(0);

 if(ServerConfig.Port == -1 || ServerConfig.MaxClients == -1 || ServerConfig.ConnectTimeout == -1)
 {
  puts("Incomplete configuration file");
  return(0);
 }
 return(1);
}

static ClientEntry *AllClients;
static GameEntry *Games;

static void en32(uint8 *buf, uint32 morp)
{
 buf[0]=morp;
 buf[1]=morp>>8;
 buf[2]=morp>>16;
 buf[3]=morp>>24;
}
    
static uint32 de32(uint8 *morp)
{
 return(morp[0]|(morp[1]<<8)|(morp[2]<<16)|(morp[3]<<24));
}

static int MakeSendTCP(ClientEntry *client, uint8 *data, uint32 len) throw(int);
static char *CleanNick(char *nick);
static int NickUnique(ClientEntry *client, const char *);
static void AddClientToGame(ClientEntry *client, uint8 id[16]) throw(int);
static void SendCommand(ClientEntry *client, uint8 cmd, uint8 *data, uint32 len) throw(int);
static void SendToAll(GameEntry *game, int cmd, uint8 *data, uint32 len) throw(int);
static void TextToClient(ClientEntry *client, const char *fmt, ...) throw(int);
static void AnnouncePlayer(ClientEntry *client, ClientEntry *newclient, bool joined) throw(int);
static void KillClient(ClientEntry *client);

#define NBTCP_LOGINLEN		0x100
#define NBTCP_LOGIN		0x200
#define NBTCP_COMMANDLEN	0x300
#define NBTCP_COMMAND		0x400

#define NBTCP_UPDATEDATA	0x800

static void StartNBTCPReceive(ClientEntry *client, uint32 type, uint32 len)
{
 client->nbtcp = (uint8 *)malloc(len);
 client->nbtcplen = len;
 client->nbtcphas = 0;
 client->nbtcptype = type;
}

static void RedoNBTCPReceive(ClientEntry *client)
{
 client->nbtcphas = 0;
}

static void EndNBTCPReceive(ClientEntry *client)
{
    free(client->nbtcp);
    client->nbtcplen = client->localplayers;
    client->nbtcphas = 0;
    client->nbtcptype = 0;
    client->nbtcp = 0;
}

static uint8 MakeMPSByte(ClientEntry *client, uint8 *mergedto)
{
 uint8 ret = 0;
 GameEntry *game = (GameEntry *)client->game;

 if(mergedto != NULL)
 {
  *mergedto = 0;

  if(client->merge_target)
   *mergedto |= 1 << (client->merge_target - 1);
 }

 for(int x = 0; x < game->TotalControllers; x++)
 {
  if(game->Clients[x] == client)
  {
   ret |= 1 << x;
  }
 }
 return(ret);
}

static char *MakeMPSString(ClientEntry *client)
{
 static char buf[1024];
 GameEntry *game = (GameEntry *)client->game;
 uint8 mps;
 uint8 mergedto;

 mps = MakeMPSByte(client, &mergedto);

 if(!mps)
 {
  buf[0] = 0;

  // FIXME, this is silly.
  for(int x = 0; x < MaxClientsPerGame; x++)
  {
   if(game->Clients[x] == client)
   {
    snprintf(buf, 256, "*Lurker %d", x + 1 - game->TotalControllers);
    break;
   }
  }
 }
 else
  snprintf(buf, 256, "*Player %s%s%s%s%s%s%s%s",
                                       (mps & 0x01) ? " 1" : "",
                                       (mps & 0x02) ? " 2" : "",
                                       (mps & 0x04) ? " 3" : "",
                                       (mps & 0x08) ? " 4" : "",
                                       (mps & 0x10) ? " 5" : "",
                                       (mps & 0x20) ? " 6" : "",
                                       (mps & 0x40) ? " 7" : "",
                                       (mps & 0x80) ? " 8" : "");


 if(client->merge_target)
 {
  snprintf(buf + strlen(buf), 768, " MT: %d", client->merge_target);
 }

 return(buf);
}

/* Returns 1 if we are back to normal game mode, 0 if more data is yet to arrive. */
static int CheckNBTCPReceive(ClientEntry *client) throw(int)
{
 if(!client->nbtcplen)
  throw(1);			/* Should not happen. */
 int l;
       
 while((l = recv(client->TCPSocket, client->nbtcp + client->nbtcphas, client->nbtcplen  - client->nbtcphas, MSG_NOSIGNAL)))
 {
  if(l == -1)
  {
   if(errno == EAGAIN || errno == EWOULDBLOCK)
    break;
   throw(1); /* Die now.  NOW. */
  }
  client->nbtcphas += l;

  //printf("Read: %d, %04x, %d, %d\n",l,client->nbtcptype,client->nbtcphas, client->nbtcplen);

  /* We're all full.  Yippie. */
  if(client->nbtcphas == client->nbtcplen)
  {
   uint32 len;

   switch(client->nbtcptype & 0xF00)
   {
    case NBTCP_UPDATEDATA:
			{
			 GameEntry *game = (GameEntry *)client->game;
		         int x, wx;

			 if(client->nbtcp[0])
			 {
			  client->pending_command = client->nbtcp[0];
			  EndNBTCPReceive(client);
			  StartNBTCPReceive(client, NBTCP_COMMANDLEN, 4);
			  return(1);
			 }

		         for(x=0,wx=0; x < game->TotalControllers; x++)
		         {
		          if(game->Clients[x] == client)
		          {
			   memcpy(&game->joybuf[game->ControllerDataOffset[x]], &client->nbtcp[1 + wx], game->ControllerDataSize[x]);
			   //uint8 tovalue;
			   //tovalue = client->nbtcp[1 + wx];
			   //game->joybuf[x] = tovalue;
		           wx += game->ControllerDataSize[x]; //++;
		          }
			 }
			 RedoNBTCPReceive(client);
			}
			return(1);
    case NBTCP_COMMANDLEN:
			{
			uint8 cmd;

			cmd = client->pending_command;

			len = de32(client->nbtcp);
			if(cmd != MDFNNPCMD_SETFPS && len > MaxCommandPayload)	/* Sanity check. */
			 throw(1);

			if(!len && !(cmd&0x80))
		        {
			 SendToAll((GameEntry*)client->game, cmd, 0, 0);
			 if(client->TCPSocket == -1) 
			 {
			  //puts("IYEEE!!!");
			  return(1);	// This client disconnected
			 }
			 EndNBTCPReceive(client);
			 StartNBTCPReceive(client,NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			}
			else if(cmd == MDFNNPCMD_SETFPS)
			{
                         ((GameEntry*)client->game)->fps = len;
                         EndNBTCPReceive(client);
                         StartNBTCPReceive(client,NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			}
                        else if(cmd == MDFNNPCMD_INTEGRITY)
                        {
                         SendToAll((GameEntry*)client->game, MDFNNPCMD_INTEGRITY, NULL, 0);
                         EndNBTCPReceive(client);
                         StartNBTCPReceive(client,NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
                        }
			else if(cmd & 0x80)
			{
                         EndNBTCPReceive(client);
			 if(len)
			  StartNBTCPReceive(client,NBTCP_COMMAND | cmd,len);
			 else	/* Woops.  Client probably tried to send a text message of 0 length. Or maybe a 0-length cheat file?  Better be safe! */
			  StartNBTCPReceive(client,NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
			}
			else throw(1);
			return(1);
			}
    case NBTCP_COMMAND: 
			{
			len = client->nbtcplen;
			uint32 tocmd = client->nbtcptype & 0xFF;

                        if(tocmd == MDFNNPCMD_INTEGRITY_RES)
                        {
                         printf("%s\n", md5_asciistr(client->nbtcp));
                        }
                        else if(tocmd == MDFNNPCMD_ECHO)
                        {
                         int newlen = len;
                         if(newlen > 256)
                          newlen = 256;

                         SendCommand(client, tocmd, client->nbtcp, newlen);
                        }
			else if(tocmd == MDFNNPCMD_SETNICK)
			{
			 int newlen = len;
			 char *newnick;

			 if(newlen > MaxNickLen)
			  newlen = MaxNickLen;

			 newnick = (char*)malloc(newlen + 1);
			 memcpy(newnick, client->nbtcp, newlen);
			 newnick[newlen] = 0;
			 newnick = CleanNick(newnick);

			 if(!newnick || NickUnique(client, newnick))
			 {
			  uint8 *ncdata;
			  uint32 ncdata_len;

                          if(!newnick)
                           newnick = strdup(MakeMPSString(client));

			  ncdata_len = strlen(client->nickname) + 1 + strlen(newnick) + 1;
			  ncdata = (uint8*)malloc(ncdata_len);

			  memcpy(ncdata, client->nickname, strlen(client->nickname));
			  ncdata[strlen(client->nickname)] = '\n';
			  strcpy((char*)ncdata + strlen(client->nickname) + 1, newnick);

			  free(client->nickname);
			  client->nickname = newnick;

                          SendToAll((GameEntry*)client->game, MDFNNPCMD_NICKCHANGED, ncdata, ncdata_len);
			  free(ncdata);
			 }
			}
			else if(tocmd == MDFNNPCMD_TEXT)
			{
                         GameEntry *game = (GameEntry *)client->game;
   
		         for(int x = 0; x < MaxClientsPerGame && game->TotalControllers; x++)
		         {
		          if(!game->Clients[x] || !game->IsUnique[x])
			   continue;

			  try
		          {
			   uint32 nicklen = strlen(client->nickname);
			   uint8 mewmewbuf[4 + nicklen + len];
			   en32(mewmewbuf, nicklen);
			   memcpy(mewmewbuf + 4, client->nickname, nicklen);
			   memcpy(mewmewbuf + 4 + nicklen, client->nbtcp, len);
			   SendCommand(game->Clients[x], tocmd, mewmewbuf, 4 + nicklen + len);
		          }
		          catch(int i)
		          {
		           KillClient(game->Clients[x]);
		          }
		         }

			 //SendToAll((GameEntry*)client->game, tocmd, (uint8 *)ma, len);
			}
			else
			 SendToAll((GameEntry*)client->game, tocmd, client->nbtcp, len);

                        EndNBTCPReceive(client);
			StartNBTCPReceive(client,NBTCP_UPDATEDATA, client->local_controllers_data_size + 1);
                        return(1);
			}
    case NBTCP_LOGINLEN:len = de32(client->nbtcp);
			if(len > (16 + 16 + 64 + 1 + /*MaxNickLen*/8192) || len < (16 + 16 + 64 + 1))
			{
			 printf("Login len exceeds maximum.\n");
			 throw(1);
			}
			EndNBTCPReceive(client);
			StartNBTCPReceive(client,NBTCP_LOGIN,len);
			return(1);
    case NBTCP_LOGIN:
			{
			 uint32 len;
			 uint8 gameid[16];
			 uint8 *sexybuf;
			 uint8 *passptr;
     			 uint8 extra[64];
			 uint32 pd_name_len = 0;

			 len = client->nbtcplen;
			 sexybuf = client->nbtcp;
     
			 /* Game ID(MD5'd game MD5 and game key on client side). */
			 memcpy(gameid, sexybuf, 16);
			 sexybuf += 16;
			 len -= 16;
     
			 passptr = sexybuf;

			 sexybuf += 16;
			 len -= 16;

			 memcpy(extra, sexybuf, 64);
			 sexybuf += 64;
			 len -= 64;

			 client->localplayers = *sexybuf;
			 client->protocol_version = extra[0];
			 client->total_controllers = extra[1];
			 pd_name_len = de32(&extra[4]);
			 client->merge_target = extra[32];

			 // Sanity check
			 if(pd_name_len > 2048)
			 {
			  puts("pd_name_len too long.");
			  throw(1);
			 }

			 // Don't even send error messages for these, because the
			 // message sending code relies on the parameters that are invalid x_x
                         if(client->total_controllers > MaxControllersPerGame)
                         {
                          puts("That number of controllers isn't supported");
                          throw(1);
                         }

			 if(!client->protocol_version)
			 {
			  puts("Protocol 0 not supported.");
			  throw(1);
		         }

                         client->controller_data_size = (int *)calloc(client->total_controllers, sizeof(int));
			 client->controller_data_offset = (int *)calloc(client->total_controllers, sizeof(int));

			 client->sendqalloced = MinSendQSize;
			 client->sendqlen = 0;
			 client->sendq = (uint8 *)calloc(client->sendqalloced, 1);

			 if(!client->controller_data_size || !client->controller_data_offset || !client->sendq)
			 {
			  puts("Memory allocation failure!");
			  throw(1);
			 }

			 if(client->protocol_version < 2)
			 {
			  for(int x = 0; x < client->total_controllers; x++)
			  {
			   client->controller_data_size[x] = 1;
			   client->controller_data_offset[x] = x;
			  }
			  client->total_controllers_data_size = client->total_controllers;
			 }
		 	 else
			 {
                          client->controller_data_size = (int *)calloc(client->total_controllers, sizeof(int));
			  int zeindex = 0;

			  for(int x = 0; x < client->total_controllers; x++)
			  {
			   client->controller_data_size[x] = extra[16 + x];
			   client->controller_data_offset[x] = zeindex;
			   zeindex += extra[16 + x];
			  }
			  client->total_controllers_data_size = zeindex;
		 	 }

			 // Hack so we can always be able to encode our command length...
			 if(client->total_controllers_data_size < 4) 
			  client->total_controllers_data_size = 4;

			 if(client->total_controllers_data_size > MaxTotalControllersDataSize)
			 {
			  // TODO: Print information about client.
			  puts("Exceeded MaxTotalControllersDataSize");
			  throw(1);
			 }

			 client->local_controllers_data_size = 0;

                         if((client->merge_target - 1) >= client->total_controllers)
                         {
			  TextToClient(client, "Invalid merge target: player %d", client->merge_target);
			  throw(1);
			 }

                         if(ServerConfig.Password)
			 {
                          if(memcmp(ServerConfig.Password,passptr,16))
                          {
                           TextToClient(client,"Invalid server password.");
                           throw(1);
                          }
			 }

			 if(client->localplayers < 1 || client->localplayers > client->total_controllers)
			 {
				puts("Invalid number of local players.");
			  TextToClient(client,"Invalid number(%d) of local players!",client->localplayers);
			  throw(1);
			 }
			 sexybuf++;
			 len -= 1;
                         AddClientToGame(client, gameid);

			 /* Get the nickname */
			 if((len - pd_name_len) > 0)
			 {
			  if((len - pd_name_len) > MaxNickLen)
			  {
			   TextToClient(client, "Nickname is too long!");
			   throw(1);
			  }
			  client->nickname = (char *)malloc(len - pd_name_len + 1);
			  memcpy(client->nickname, sexybuf, len - pd_name_len);
			  client->nickname[len - pd_name_len] = 0;
			  if((client->nickname = CleanNick(client->nickname)))
			   if(!NickUnique(client, client->nickname))       /* Nickname already exists */
			   {
		            free(client->nickname);
		            client->nickname = 0;
			   }
		          }
			 char *mps = MakeMPSString(client);

			 if(!client->nickname)
			  client->nickname = strdup(mps);

			 printf("Client(protocol %d) %d assigned to game %d as player %s <%s>\n", client->protocol_version, client->id, (int)((GameEntry*)client->game - Games), mps, client->nickname);

			 int x;
			 GameEntry *tg=(GameEntry *)client->game;

			 for(x = 0; x < MaxClientsPerGame && tg->TotalControllers; x++)
			  if(tg->Clients[x] && tg->IsUnique[x])
			  {
			   if(tg->Clients[x] != client)		  
			   {
			    try
			    {
			     AnnouncePlayer(tg->Clients[x], client, 1);
			    } catch(int i)
			    {
			     KillClient(tg->Clients[x]);
			    }
			    AnnouncePlayer(client, tg->Clients[x], 1);
			   }
			   else
			   {
			    AnnouncePlayer(client, client, 1);
			   }
			  }
			} 
			EndNBTCPReceive(client);
			StartNBTCPReceive(client,NBTCP_UPDATEDATA,client->local_controllers_data_size + 1);
			return(1);
   }
  }
 }
 return(0);
}

int ListenSocket;

// Remove any control characters or reserved characters from the nickname.
static char *CleanNick(char *nick)
{
 for(unsigned int x = 0; x < strlen(nick); x++)
  if(nick[x] == '<' || nick[x] == '>' || nick[x] == '*' || (nick[x] < 0x20))
   nick[x] = 0;

 if(!strlen(nick))
 {
  free(nick);
  nick = 0;
 }

 return(nick);
}

static int NickUnique(ClientEntry *client, const char *newnick)
{
 int x;
 GameEntry *game = (GameEntry *)client-> game;

 for(x = 0; x < MaxClientsPerGame; x++)
  if(game->Clients[x] && client != game->Clients[x])
   if(!strcasecmp(newnick, game->Clients[x]->nickname))
    return(0);
 return(1);
}

static int MakeSendTCP(ClientEntry *client, uint8 *data, uint32 len) throw(int)
{
 // If there's no data in our software sendq, try sending this data without using the soft sendq.
 if(!client->sendqlen)
 {
  int32 sent = send(client->TCPSocket, data, len, MSG_NOSIGNAL);

  if(sent < 0)
  {
   if(errno == EAGAIN || errno == EWOULDBLOCK)
    sent = 0;
   else
   {
    //puts("Fatal error on send()");
    throw(1);
   }
  }

  assert(sent <= len);

  // ...and adjust the data pointer and length appropriately.
  data += sent;
  len -= sent;

  // If we sent all the data, no need to go through the software sendq code at all~
  if(!len)
   return(1);
 }




 assert(client->sendq);

 
 if((client->sendqlen + len) > client->sendqalloced)
 {
  uint32 new_alloc_size;

  new_alloc_size = client->sendqlen + len;

  if(new_alloc_size > MaxSendQSize)
  {
   // TODO: More information about client and game client belongs to.
   printf("Dropped client due to exceeding MaxSendQSize by %d bytes\n", MaxSendQSize - new_alloc_size);
   throw(1);
  }

  new_alloc_size += new_alloc_size >> 1;
  if(new_alloc_size > MaxSendQSize)
   new_alloc_size = MaxSendQSize;

  if(!(client->sendq = (uint8 *)realloc(client->sendq, new_alloc_size)))
  {
   printf("Dropped client due to realloc failure: %m\n");
   throw(1);
  }
  client->sendqalloced = new_alloc_size;
 }

 //puts("SendQ path");

 // Append new data to the end of the sendq
 memcpy(client->sendq + client->sendqlen, data, len);
 client->sendqlen += len;

 // Now try to send all the data in the sendq
 int32 sent = send(client->TCPSocket, client->sendq, client->sendqlen, MSG_NOSIGNAL);

 if(sent < 0) 
 {
  if(errno == EAGAIN || errno == EWOULDBLOCK)
   sent = 0;
  else
  {
   //puts("Fatal error on send()");
   throw(1);
  }
 }

 if(sent)
 {
  uint32 new_alloc_size;

  client->sendqlen -= sent;
  memmove(client->sendq, client->sendq + sent, client->sendqlen);


  new_alloc_size = client->sendqlen;
  if(new_alloc_size < MinSendQSize)
   new_alloc_size = MinSendQSize;

  if(!(client->sendq = (uint8 *)realloc(client->sendq, new_alloc_size)))
  {
   printf("Dropped client due to realloc failure: %m\n");
   throw(1);
  }
  client->sendqalloced = new_alloc_size;
 }

 return(1);
}

static void SendCommand(ClientEntry *client, uint8 cmd, uint8 *data, uint32 len) throw(int)
{
 uint8 poo[client->total_controllers_data_size + 1];

 memset(poo, 0, client->total_controllers_data_size);

 poo[client->total_controllers_data_size] = cmd;

 if((cmd & 0x80) && data != NULL)
  en32(poo, len);

 MakeSendTCP(client, poo, client->total_controllers_data_size + 1);

 if((cmd & 0x80) && data != NULL)
  MakeSendTCP(client, data, len);
}

static void SendToAll(GameEntry *game, int cmd, uint8 *data, uint32 len) throw(int)
{
 for(int x = 0; x < MaxClientsPerGame && game->TotalControllers; x++)
 {
  if(!game->Clients[x] || !game->IsUnique[x]) continue;
   
  try
  {
   SendCommand(game->Clients[x], cmd, data, len);
  } 
  catch(int i)
  {
   KillClient(game->Clients[x]);
  }
 } 
}

static void TextToClient(ClientEntry *client, const char *fmt, ...) throw(int)
{
 uint8 moo[4 + 1024]; // nick len(=0) and message
 va_list ap;

 va_start(ap,fmt);
 vsnprintf((char*)moo + 4, 1024, fmt, ap);
 va_end(ap);
 
 en32(moo, 0); // nick len of 0 to signify this is a server message
 SendCommand(client, 0x90, moo, 4 + strlen((char*)moo + 4));
}

static void AnnouncePlayer(ClientEntry *client, ClientEntry *newclient, bool joined) throw(int)
{
 uint32 datalen = 1 + 1 + strlen(newclient->nickname);
 uint8 data[datalen];

 data[0] = MakeMPSByte(newclient, &data[1]);
 memcpy(data + 2, newclient->nickname, strlen(newclient->nickname));
 if(client == newclient)
  SendCommand(client, joined ? MDFNNPCMD_YOUJOINED : MDFNNPCMD_YOULEFT, data, datalen);
 else
  SendCommand(client, joined ? MDFNNPCMD_PLAYERJOINED : MDFNNPCMD_PLAYERLEFT, data, datalen);

 if(client == newclient)
 {
  uint8 localmask = MakeMPSByte(newclient, NULL);
  client->local_controllers_data_size = 0;

  for(int x = 0; x < client->total_controllers; x++)
   if(localmask & (1 << x))
    client->local_controllers_data_size += client->controller_data_size[x];
 }
}

static void KillClient(ClientEntry *client)
{
 GameEntry *game;

 if(!client)
 {
  puts("Erhm, attempt to kill NULL client o_O");
  return;
 }
 if(client->TCPSocket == -1)
 {
  puts("Attempt to kill client multiple times.  The grim reaper has gone mad!");
  return;
 }

 game = (GameEntry *)client->game;
 if(game)
 {
  // Remove the client from the game's player list BEFORE we do an announcement that the client has left.  Which should be obvious, but we used to do it backwards
  // and wonkiness randomly ensued. :/
  //
  // And uncomment out this code immediately below if we ever have a "nice" quit handshake rather than a hard disconnect.
  // try
  // {
  //  AnnouncePlayer(client, client, 0);
  // }
  // catch(int i)
  // {
  //  // Don't do anything here, since we're already on the way to kill the client.  KILL IT DEAD.
  // }

  for(int x = 0; x < MaxClientsPerGame; x++)
   if(game->Clients[x])
    if(game->Clients[x] == client)
     game->Clients[x] = NULL;

  // Now announce to the remaining players that the client has left.
  for(int x = 0; x < MaxClientsPerGame && game->TotalControllers; x++)
  {
   if(!game->Clients[x] || !game->IsUnique[x])
    continue;

   try
   {
    AnnouncePlayer(game->Clients[x], client, 0);
   }
   catch(int i)
   {
    KillClient(game->Clients[x]);
   }
  }

  time_t curtime = time(0);
  printf("Player <%s> disconnected from game %d on %s", client->nickname, (int)(game - Games), ctime(&curtime));


  if(game->TotalControllers)	// Make sure the game still exists to destroy, since we could have killed it in a recursive call to KillClient() above.
  {
   bool AnyClientsLeft = false;		// Any clients left connected to this game?

   for(int x = 0; x < MaxClientsPerGame && !AnyClientsLeft; x++)
    if(game->Clients[x])
     AnyClientsLeft = true;

   if(!AnyClientsLeft)
   {
    printf("Game %d destroyed.\n", (int)(game-Games));
    memset(game, 0, sizeof(GameEntry));
    game = 0;
   }
  }
 }
 else
 {
  time_t curtime = time(0);
  printf("Unassigned client %d disconnected on %s",client->id, ctime(&curtime));
 }

 if(client->controller_data_size)
  free(client->controller_data_size);

 if(client->controller_data_offset)
  free(client->controller_data_offset);

 if(client->sendq)
  free(client->sendq);

 if(client->nbtcp)
  free(client->nbtcp);
 
 if(client->nickname) 
  free(client->nickname);

 if(client->TCPSocket != -1)
  close(client->TCPSocket);
 memset(client, 0, sizeof(ClientEntry));
 client->TCPSocket = -1;
}

static void AddClientToGame(ClientEntry *client, uint8 id[16]) throw(int)
{
 int wg;
 GameEntry *game,*fegame;

 retry:

 game = NULL;
 fegame = NULL;

 /* First, find an available game. */
 for(wg=0; wg<ServerConfig.MaxClients; wg++)
  if(!Games[wg].TotalControllers && !fegame)
  {
   if(!fegame)
    fegame=&Games[wg];
  }
  else if(Games[wg].TotalControllers && !memcmp(Games[wg].id,id,16)) /* A match was found! */
  {
   game = &Games[wg];
   break;
  }

 if(game)
 {
  if(game->ProtocolVersion != client->protocol_version)
  {
   puts("MOO");
   TextToClient(client, "Protocol version mismatch.  Game: %d, You: %d", game->ProtocolVersion, client->protocol_version);
   throw(1);
  }
  if(game->TotalControllers != client->total_controllers || game->TotalControllersDataSize != client->total_controllers_data_size)
  {
   TextToClient(client, "Hunh?  Controllers mismatch...");
  }
 }
 if(!game) /* Hmm, no game found.  Guess we'll have to create one. */
 {
  game=fegame;
  printf("Game %d added\n",(int)(game-Games));
  memset(game, 0, sizeof(GameEntry));

  game->TotalControllers = client->total_controllers; // total_controllers is validated in the user login code
  game->TotalControllersDataSize = client->total_controllers_data_size;

  memcpy(game->ControllerDataSize, client->controller_data_size, game->TotalControllers * sizeof(int));
  memcpy(game->ControllerDataOffset, client->controller_data_offset, game->TotalControllers * sizeof(int));

  game->fps = 1008307711; // NES NTSC FPS magical number from beyond the moon, the default.
  game->last_time = MBL_Time64();
  game->ProtocolVersion = client->protocol_version;
  game->Clients = (ClientEntry **)calloc(MaxClientsPerGame, sizeof(ClientEntry *));

  memcpy(game->id, id, 16);
 }


 int n;

 for(n = 0; n < MaxClientsPerGame; n++)
  if(game->Clients[n])
  {
   try
   {
    SendCommand(game->Clients[n], 0x81, NULL, 0);
    break;
   }
   catch(int i)
   {
    KillClient(game->Clients[n]);
    /* All right, now the client we sent the request to is killed.  I HOPE YOU'RE HAPPY! */
    /* Since killing a client can destroy a game, we'll need to see if the game was trashed.  
       If it was, then "goto retry", and try again.  Ugly, yes.  I LIKE UGLY.
    */
    if(!game->TotalControllers)
     goto retry;
   }
  }

 int instancecount = client->localplayers;

 for(n = 0; n < MaxClientsPerGame; n++)
 {
  if(!game->Clients[n])
  {
   game->Clients[n] = client;
   if(instancecount == client->localplayers)
    game->IsUnique[n] = 1;
   else
    game->IsUnique[n] = 0;
   instancecount --;
   if(!instancecount)
    break;
  }
 }

 /* Game is full. */
 if(n == MaxClientsPerGame)
 {
  for(n = 0; n < MaxClientsPerGame; n++)
   if(game->Clients[n] == client)
    game->Clients[n] = 0;

  TextToClient(client, "Sorry, game is full.  %d instance(s) tried, %d available.", client->localplayers, client->localplayers - instancecount);
  throw(1);
 }

 client->game = (void *)game;
}

//#define MDFNSERVER_IPV6

int main(int argc, char *argv[])
{
 #ifdef MDFNSERVER_IPV6
 struct sockaddr_in6 sockin;
 #else
 struct sockaddr_in sockin;
 #endif

 socklen_t sockin_len;

 if(argc < 2)
 {
  if(strchr(argv[0],' '))
   printf("Usage: \"%s\" <configfile>\n",argv[0]);
  else
   printf("Usage: %s <configfile>\n",argv[0]);
  exit(-1);
 }

 printf("Starting Mednafen-Server %s\n\n", VERSION);

 if(!LoadConfig(argv[1]))
 {
  puts("Error loading configuration file");
  exit(-1);
 }

 Games = (GameEntry *)malloc(sizeof(GameEntry) * ServerConfig.MaxClients);
 AllClients = (ClientEntry *)malloc(sizeof(ClientEntry) * ServerConfig.MaxClients);

 memset(Games, 0, sizeof(GameEntry) * ServerConfig.MaxClients);
 memset(AllClients, 0, sizeof(ClientEntry) * ServerConfig.MaxClients);

 for(int x = 0; x < ServerConfig.MaxClients; x++)
 {
  AllClients[x].TCPSocket = -1;
 }

 /* First, we need to create a socket to listen on. */
 #ifdef MDFNSERVER_IPV6
 ListenSocket = socket(AF_INET6, SOCK_STREAM, 0);
 #else
 ListenSocket = socket(AF_INET, SOCK_STREAM, 0);
 #endif

 /* Set send buffer size to 262,144 bytes. */
 int sndbufsize = 262144;
 if(setsockopt(ListenSocket, SOL_SOCKET, SO_SNDBUF, &sndbufsize, sizeof(int)))
  puts("Send buffer size set failed.");

 int tcpopt = 1;
 if(setsockopt(ListenSocket, SOL_TCP, TCP_NODELAY, &tcpopt, sizeof(int)))
 {
  puts("Nodelay fail");
  exit(-1);
 }

 memset(&sockin, 0, sizeof(sockin));

#ifdef MDFNSERVER_IPV6
 memcpy(&sockin.sin6_addr, &in6addr_any, sizeof(in6_addr));	// = IN6ADDR_ANY_INIT;
 sockin.sin6_family = AF_INET6;
 sockin.sin6_port = htons(ServerConfig.Port);
#else
 sockin.sin_addr.s_addr = INADDR_ANY;
 sockin.sin_family = AF_INET;
 sockin.sin_port = htons(ServerConfig.Port);
#endif
 sockin_len = sizeof(sockin);

 if(bind(ListenSocket, (struct sockaddr *)&sockin, sockin_len))
 {
  printf("%s\n", strerror(errno));
  puts("Bind error.");
  exit(-1);
 }
 if(listen(ListenSocket, 8))
 {
  puts("Listen error.");
  exit(-1);
 }
 printf("Listening on port %d.\n",ServerConfig.Port);

 /* We don't want to block on accept() */
 fcntl(ListenSocket, F_SETFL, fcntl(ListenSocket, F_GETFL) | O_NONBLOCK);

 /* Now for the BIG LOOP. */
 while(1)
 {
  int n;

  for(n=0; n<ServerConfig.MaxClients; n++)
  {
   if(AllClients[n].TCPSocket != -1) continue;
   if((AllClients[n].TCPSocket = accept(ListenSocket, (struct sockaddr *)&sockin, &sockin_len)) != -1)
   {
    /* We have a new client.  Yippie. */

    fcntl(AllClients[n].TCPSocket, F_SETFL, fcntl(AllClients[n].TCPSocket, F_GETFL) | O_NONBLOCK);

    AllClients[n].timeconnect = time(0);
    AllClients[n].id = n;

#ifdef MDFNSERVER_IPV6
    {
     char tbuf[256];
     printf("Client %d connecting from %s on %s", n, inet_ntop(AF_INET6, &sockin.sin6_addr, tbuf, sizeof(tbuf)), ctime(&AllClients[n].timeconnect));
    }
#else
    printf("Client %d connecting from %s on %s", n, inet_ntoa(sockin.sin_addr), ctime(&AllClients[n].timeconnect));
#endif
    StartNBTCPReceive(&AllClients[n], NBTCP_LOGINLEN, 4);
   }
  }  

  /* Check for users still in the login process(not yet assigned a game). BOING */
  time_t curtime = time(0);
  for(n = 0; n < ServerConfig.MaxClients; n++)
   try
   {
    if(AllClients[n].TCPSocket != -1 && !AllClients[n].game)
    {
     if((AllClients[n].timeconnect + ServerConfig.ConnectTimeout) < curtime)
     {
      KillClient(&AllClients[n]);
     }
     else
      while(CheckNBTCPReceive(&AllClients[n])) {};
    }
   }
   catch(int i) 
   {
    KillClient(&AllClients[n]);
   }
  int whichgame;
  for(whichgame = 0; whichgame < ServerConfig.MaxClients; whichgame ++)
  {
   if(!Games[whichgame].fps)
    continue;

   int64 last_time_inc = (int64)65536 * 256 * 1000000 / Games[whichgame].fps;

   if((Games[whichgame].last_time + last_time_inc) > MBL_Time64())
    continue;
   Games[whichgame].last_time += last_time_inc;
   //printf("%lld\n", last_time_inc);

   /* Now, the loop to get data from each client.  Meep. */
   for(n = 0; n < MaxClientsPerGame && Games[whichgame].TotalControllers; n++)
   {
    try
    {
     ClientEntry *client = Games[whichgame].Clients[n];
     if(!client || !Games[whichgame].IsUnique[n]) 
      continue;

     while(CheckNBTCPReceive(client)) {};
    } // catch
    catch(int i)
    {
     KillClient(Games[whichgame].Clients[n]);
    }
   } // A games clients

   if(!Games[whichgame].fps) // Oops, client exiting killed it.
   {
    continue;
   }

   uint8 merge_data_buf[Games[whichgame].TotalControllersDataSize + 1];

   memcpy(merge_data_buf, Games[whichgame].joybuf, sizeof(merge_data_buf));

   /* Now merge!!! */
   for(n = 0; n < Games[whichgame].TotalControllers; n++)
   {
    if(!Games[whichgame].Clients[n] || !Games[whichgame].IsUnique[n]) 
     continue;

    if(Games[whichgame].Clients[n]->merge_target)
    {
     int offshoo = Games[whichgame].ControllerDataOffset[n];
     int offshoomt = Games[whichgame].ControllerDataOffset[ Games[whichgame].Clients[n]->merge_target - 1 ];

     for(int zoom = 0; zoom < Games[whichgame].ControllerDataSize[n]; zoom++)
     {
      merge_data_buf[offshoomt + zoom] |= Games[whichgame].joybuf[offshoo + zoom];
     }
    }
   }

   /* Now we send the data to all the clients. */
   for(n = 0; n < MaxClientsPerGame && Games[whichgame].TotalControllers; n++)
   {
    if(!Games[whichgame].Clients[n] || !Games[whichgame].IsUnique[n])
     continue;

    try
    {
     MakeSendTCP(Games[whichgame].Clients[n], merge_data_buf, sizeof(merge_data_buf));
    }
    catch(int i)
    {
     KillClient(Games[whichgame].Clients[n]);
    }
   } // A game's clients
  } // Games

  // Now, determine the amount of time we should sleep, and sleep it.
  int64 min_sleep = 25000;
  int64 curgametime = MBL_Time64();
  for(whichgame = 0; whichgame < ServerConfig.MaxClients; whichgame ++)
  {
   if(!Games[whichgame].fps)
    continue;
   int64 last_time_inc = (int64)65536 * 256 * 1000000 / Games[whichgame].fps;
   int64 sandwich = (Games[whichgame].last_time + last_time_inc) - curgametime;
   if(sandwich <= 0)
   {
    min_sleep = 0;
    break;
   }
   else if(sandwich < min_sleep) min_sleep = sandwich;
  }
  if(min_sleep > 0)
   MBL_Sleep64(min_sleep);
  //printf("%lld, %lld\n", min_sleep, MBL_Time64() - curgametime);
 } // while(1)
}
